#include <SPI.h>
#include <SdFat.h>

#include <Adafruit_GFX.h>
#include <Adafruit_ST7735.h>

#include <flash_stm32.h>
#include <libmaple/scb.h>

#include <USBComposite.h>

#define PRODUCT_ID 0x29

#define SCS       0xE000E000
#define SCB       (SCS + 0xD00)
#define SCB_VTOR  (SCB + 0x08)

#define FL_START_ADDR  0x08008000      // FLASH_ErasePage
#define FL_PRG_SIZE        98304

#define TFT_CS     PB12
#define TFT_RST    PB10
#define TFT_DC     PB5
#define D_BR       PB8
#define SD_CS      PB9

#define bal        PB1
#define jobb       PB0

#define ps_enb     PA14
#define ps_btn     PB11

#define color24bit(color) (((color & 0xF80000) >> 8) & ((color & 0x00FC00) >> 5) & ((color & 0x0000F8) >> 3))

USBMassStorage MassStorage;
bool storageEnabled = false;
uint32 cardSize;

SPIClass TFTSPI(2);
Adafruit_ST7735 tft = Adafruit_ST7735(&TFTSPI, TFT_CS,  TFT_DC, TFT_RST);

File root;
File entry;
char filename[30];
char selected[30];

SdFat SD;
const uint8_t spiSpeed = SPI_QUARTER_SPEED;

uint32_t cache = 0;
int filesize = 0;

bool sdwrite(const uint8_t *writebuff, uint32_t startSector, uint16_t numSectors) {
  return SD.card()->writeBlocks(startSector, writebuff, numSectors);
}

bool sdread(uint8_t *readbuff, uint32_t startSector, uint16_t numSectors) {
  return SD.card()->readBlocks(startSector, readbuff, numSectors);
}

void initReader() {
  cardSize = SD.card()->cardSize();
  MassStorage.setDriveData(0, cardSize, sdread, sdwrite);
  MassStorage.registerComponent();
  USBComposite.begin();
  storageEnabled = true;
}

void  setMspAndJump(uint32_t usrAddr) {
  typedef void(*funcPtr)(void);
  uint32_t jumpAddr = *(volatile uint32_t *)(usrAddr + 0x04);
  funcPtr usrMain = (funcPtr) jumpAddr;
  SCB_BASE->VTOR = (volatile uint32_t) (usrAddr);
  asm volatile("msr msp, %0"::"g"(*(volatile uint32_t *)usrAddr));
  usrMain();
}

void setup() {
  afio_cfg_debug_ports(AFIO_DEBUG_NONE);
  pinMode(D_BR, PWM);
  pinMode(bal, INPUT_ANALOG);
  pinMode(jobb, INPUT_ANALOG);
  pinMode(ps_btn, INPUT_PULLDOWN);
  pinMode(ps_enb, OUTPUT);
  delay(1000);
  digitalWrite(ps_enb, HIGH);

  USBComposite.setProductId(PRODUCT_ID);
  analogWrite(D_BR, 60);
  tft.initR(INITR_BLACKTAB);
  tft.setRotation(1);
  tft.fillScreen(ST77XX_BLACK);
  tft.setCursor(0, 0);
  tft.setTextColor(ST77XX_WHITE, ST77XX_BLACK);
  tft.setTextSize(1);
  tft.setTextWrap(true);
  if (!SD.begin(SD_CS, spiSpeed)) {
    tft.println("SD HIBA");
    while (!SD.begin(SD_CS, spiSpeed)) {
      if (digitalRead(ps_btn)) {
        delay(1000);
        if (digitalRead(ps_btn)) {
          digitalWrite(ps_enb, LOW);
        }
      }
      delay(500);
    }
  }
  initReader();
}

int selectedFile = 0;
int i = 0;

void loop() {
  if (storageEnabled) {
    int updt = millis();
    while ((millis() - updt) < 1000) {
      MassStorage.loop();
    }
  } else {
    delay(1000);
  }
  tft.setTextColor(ST77XX_WHITE, ST77XX_BLACK);
  tft.setCursor(0, 0);
  tft.println("Kovetkezo:B\nKivalasztas:J\nElozo program futtatasa:JH");
  root = SD.open("/");
  i = 0;
  while (true) {
    entry =  root.openNextFile();
    if (! entry) {
      break;
    }
    entry.getName(filename, 30);
    if ((!entry.isHidden()) && (!entry.isSystem()) && (!entry.isDir())) {
      if (i == selectedFile) {
        tft.setTextColor(ST77XX_WHITE, ST77XX_BLUE);
        strcpy(selected, filename);
      }      
      tft.println(filename);
      tft.setTextColor(ST77XX_WHITE, ST77XX_BLACK);
      i++;
    }
  }
  if (analogRead(bal) < 1000) {
    delay(1000);
    if (analogRead(bal) < 1000) {
      if (selectedFile == (i - 1)) {
        selectedFile = 0;
      } else {
        selectedFile++;
      }
    }
  }
  if (digitalRead(ps_btn)) {
    delay(1000);
    if (digitalRead(ps_btn)) {
      digitalWrite(ps_enb, LOW);
    }
  }
  if (analogRead(jobb) > 3000) {
    delay(2000);
    if (analogRead(jobb) > 3000) {
      setMspAndJump(FL_START_ADDR);
    }
  }
  if (analogRead(jobb) < 1000) {
    delay(2000);
    if (analogRead(jobb) < 1000) {
      entry = SD.open(selected);
      tft.println("Masolas folyamatban");
      filesize = entry.size();
      FLASH_Unlock();
      for (int j = 0; j <= (filesize / 1024); j++ ) {
        FLASH_ErasePage(FL_START_ADDR + j * 1024);
      }
      tft.println("Memoria torolve");
      for (int j = 0; j < (filesize / 4); j++) {
        entry.read(&cache, 4);
        FLASH_ProgramHalfWord(FL_START_ADDR + j * 4, ((uint16_t)(cache & 0x0000FFFF)));
        FLASH_ProgramHalfWord(FL_START_ADDR + j * 4 + 2, ((uint16_t)((cache & 0xFFFF0000) >> 16)));
      }
      FLASH_Lock();
      tft.println("Program inditasa");
      entry.close();
      setMspAndJump(FL_START_ADDR);
    }
  }
  entry.close();
}
