#include <SPI.h>

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library for ST7735

#define TFT_CS     PB12
#define TFT_RST    PB10
#define TFT_DC     PB5
#define D_BR       PB8
#define SD_CS      PB9

#define X1         PA4
#define X2         PA3
#define Y1         PA2
#define Y2         PA1

#define bal        PB1
#define jobb       PB0

#define ps_enb     PA14
#define ps_btn     PB11

#define color24bit(color) (((color & 0xF80000) >> 8) | ((color & 0x00FC00) >> 5) | ((color & 0x0000F8) >> 3))

SPIClass TFTSPI(2);
Adafruit_ST7735 tft = Adafruit_ST7735(&TFTSPI, TFT_CS,  TFT_DC, TFT_RST);

const uint8_t spiSpeed = SPI_CLOCK_DIV2; // SPI_HALF_SPEED;

int board[8][14];
const int colors[] = {ST77XX_RED, ST77XX_BLUE, ST77XX_GREEN, ST77XX_CYAN, ST77XX_MAGENTA, ST77XX_YELLOW, 0xFBE0}; // last is orange
const int shapeTypes[7][4][2] = {
  {{0, 0}, {1, 1}, {1, 1}, {0, 0}},
  {{0, 0}, {0, 1}, {0, 1}, {1, 1}},
  {{1, 0}, {1, 1}, {0, 1}, {0, 0}},
  {{0, 1}, {1, 1}, {1, 0}, {0, 0}},
  {{1, 0}, {1, 0}, {1, 0}, {1, 0}},
  {{0, 0}, {1, 0}, {1, 1}, {1, 0}},
  {{0, 0}, {1, 0}, {1, 0}, {1, 1}}
};
int shapes[255];
bool falling[255];
int shapeCount = 0;
int lastUpdate = 0;

bool ingame = false;
bool started = false;

int getMin(int a, int b) {
  if (a <= b) {
    return a;
  } else {
    return b;
  }
}

int getMax(int a, int b) {
  if (a >= b) {
    return a;
  } else {
    return b;
  }
}

void drawCube (int x, int y, int color) {
  tft.fillRect((40 + 10 * x), (117 - 10 * y), 10, 10, color);
  tft.drawRect((40 + 10 * x), (117 - 10 * y), 10, 10, color24bit(0x7F7F7F));
}

void updateScreen () {
  int shape = 0;
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 12; j++) {
      shape = board[i][j];
      if (shape == 0) {
        tft.fillRect((40 + 10 * i), (117 - 10 * j), 10, 10, ST77XX_BLACK);
      } else {
        drawCube(i, j, shapes[shape]);
      }
    }
  }
}

void dropShapes() {
  for (int s = 1; s <= shapeCount; s++) {
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 14; j++) {
        if (((board[i][j] == s) && (((board[i][j - 1] != 0) && (board[i][j - 1] != s)) || (j == 0)))) {
          falling[s] = false;
        }
      }
    }
    if (falling[s]) {
      for (int i = 0; i < 8; i++) {
        for (int j = 0; j < 14; j++) {
          if (board[i][j] == s) {
            board[i][j] = 0;
            board[i][j - 1] = s;
          }
        }
      }
    }
  }
}

void addNewShape() {
  falling[shapeCount] = true;
  int shape = rand() % 7;
  shapes[shapeCount] = colors[shape];
  int xOfs = rand() % 5;
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 2; j++) {
      board[0 + i + xOfs][12 + j] = (shapeTypes[shape][i][j] * shapeCount);
    }
  }
}

void rotateShape() {
  float orig[2];
  int coords [4][2];
  int found = 0;
  int temp = 0;
  bool itfits = true;

  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 14; j++) {
      if (board[i][j] == shapeCount) {
        coords [found][0] = i;
        coords [found][1] = j;
        found ++;
      }
      if (found == 4) {
        break;
      }
    }
    if (found == 4) {
      break;
    }
  }

  orig[0] = float(getMin(getMin(coords [0][0], coords [1][0]), getMin(coords [2][0], coords [3][0])) + getMax(getMax(coords [0][0], coords [1][0]), getMax(coords [2][0], coords [3][0]))) / 2;
  orig[1] = float(getMin(getMin(coords [0][1], coords [1][1]), getMin(coords [2][1], coords [3][1])) + getMax(getMax(coords [0][1], coords [1][1]), getMax(coords [2][1], coords [3][1]))) / 2;
  for (int i = 0; i < 4; i++) {
    temp = coords[i][0];
    coords[i][0] = int (-1 * float(coords[i][1] - orig[1]) + orig[0] + 0.5f);
    coords[i][1] = int (float(temp -  orig[0]) + orig[1] + 0.5f) ;
  }
  for (int i = 0; i < 4; i++) {
    if (!(coords[i][0] >= 0) or !(coords[i][0] < 8) or  !(coords[i][1] >= 0)) {
      itfits = false;
    } else if ((board[coords[i][0]][coords[i][1]] != 0) && (board[coords[i][0]][coords[i][1]] != shapeCount)) {
      itfits = false;
    }

  }
  if (itfits) {
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 14; j++) {
        if (board[i][j] == shapeCount) {
          board[i][j] = 0;
        }
      }
    }
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 14; j++) {
        for (int c = 0; c < 4; c++) {
          if ((i == coords[c][0]) && (j == coords[c][1])) {
            board[i][j] = shapeCount;
          }
        }
      }
    }
  }
}

void moveShape(int steps) {
  int coords [4][2];
  int found = 0;
  bool itfits = true;

  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 14; j++) {
      if (board[i][j] == shapeCount) {
        coords [found][0] = i;
        coords [found][1] = j;
        found ++;
        board[i][j] = 0;
      }
      if (found == 4) {
        break;
      }
    }
    if (found == 4) {
      break;
    }
  }
  for (int i = 0; i < 4; i++) {
    if (((coords[i][0] + steps) < 0) || ((coords[i][0] + steps > 7))) {
      itfits = false;
    } else if ((board[coords[i][0] + steps][coords[i][1]] != 0) && (board[coords[i][0] + steps][coords[i][1]] != shapeCount)) {
      itfits = false;
    }
  }
  if (itfits) {
    for (int i = 0; i < 4; i++) {
      coords[i][0] += steps;
    }
  }
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 14; j++) {
      for (int c = 0; c < 4; c++) {
        if ((i == coords[c][0]) && (j == coords[c][1])) {
          board[i][j] = shapeCount;
        }
      }
    }
  }
}

void checkLines() {
  bool full = true;
  for (int j = 0; j < 12; j++) {
    full = true;
    for (int i = 0; i < 8; i++) {
      if (board[i][j] == 0) {
        full = false;
      }
    }
    if (full) {
      for (int i = 0; i < 8; i++) {
        for (int y = j; y < 12; y++) {
          board[i][y] = board[i][y + 1];
        }
        board[i][12] = 0;
      }
    }
  }
}

void setup() {
  afio_cfg_debug_ports(AFIO_DEBUG_NONE);
  pinMode(ps_btn, INPUT_PULLDOWN);
  pinMode(ps_enb, OUTPUT);
  digitalWrite(ps_enb, HIGH);
  pinMode(D_BR, PWM);
  pinMode(bal, INPUT_ANALOG);
  pinMode(jobb, INPUT_ANALOG);
  pinMode(Y2, INPUT_ANALOG);
  pinMode(X2, INPUT_ANALOG);
  pinMode(Y1, INPUT_ANALOG);
  pinMode(X1, INPUT_ANALOG);
  analogWrite(D_BR, 120);

  tft.initR(INITR_BLACKTAB);
  tft.setRotation(1);
  tft.fillScreen(ST77XX_BLACK);
  tft.setCursor(0, 0);
  tft.setTextColor(ST77XX_WHITE, ST77XX_BLACK);//col24to16bit(0x808080));
  tft.setTextSize(1);
  tft.setTextWrap(true);
  tft.setTextColor(ST77XX_WHITE, ST77XX_BLACK);
  randomSeed((analogRead(Y2) + analogRead(X2) - 4000) * 20);
  tft.fillScreen(ST77XX_BLACK);
  tft.setCursor(0, 0);
  tft.println("TETRIS");
  tft.println("Jatek inditasa: J");
}

void loop() {
  if (!ingame) {
    if (started) {
      tft.fillScreen(ST77XX_BLACK);
      tft.setCursor(0, 0);      
      tft.println("TETRIS");
      tft.println("Jatek vege");
      tft.println("Pontszam: " + String(shapeCount));
      tft.println("Jatek inditasa: J");
      started = false;
    }
    if (digitalRead(ps_btn)) {
      delay(1000);
      if (digitalRead(ps_btn)) {
        digitalWrite(ps_enb, LOW);
      }
    }
    if (analogRead(jobb) < 1000) {
      delay(1000);
      if (analogRead(jobb) < 1000) {
        ingame = true;
        started = true;
        shapeCount = 0;
        for (int i = 0; i < 8; i++) {
          for (int j = 0; j < 14; j++) {
            board[i][j] = 0;
          }
        }
        tft.fillScreen(ST77XX_BLACK);
        tft.drawFastVLine(39, 0, 127, 0xFFFFFF);
        tft.drawFastVLine(120, 0, 127, 0xFFFFFF);
        delay(2000);
        lastUpdate = millis();
      }
    }
  }
  if (ingame) {
    if (digitalRead(ps_btn)) {
      delay(1000);
      if (digitalRead(ps_btn)) {
        digitalWrite(ps_enb, LOW);
      }
    }
    if ((millis() - lastUpdate) > 200) {
      updateScreen();
      if (!falling[shapeCount]) {
        for (int i = 0; i < 8; i++) {
          if (board[i][12] == shapeCount && shapeCount != 0) {
            ingame = false;
          }
        }
        if (ingame) {
          shapeCount++;
          addNewShape();
        }
      }
      tft.setCursor(125, 10);
      tft.println(shapeCount);
      tft.println("Leejt\nB\n\nForgat\nJ");
      if (falling[shapeCount]) {
        if (analogRead(jobb) < 1000) {
          rotateShape();
          delay(200);
          lastUpdate = millis();
        } else if (analogRead(bal) < 1000) {
          while (falling[shapeCount]) {
            dropShapes();
          }
          delay(200);
          lastUpdate = millis();
        } else if (analogRead(X2) > 2700) {
          moveShape(1);
          delay(200);
          lastUpdate = millis();
        } else if (analogRead(X2) < 1600) {
          moveShape(-1);
          delay(200);
          lastUpdate = millis();
        }
      }
    }
    if ((millis() - lastUpdate) > 1000) {
      dropShapes();
      checkLines();
      lastUpdate = millis();
    }
  }
}
